use preserves::error::{is_eof_io_error, is_syntax_io_error};
use preserves::symbol::Symbol;
use preserves::value::de::from_value as deserialize_from_value;
use preserves::value::decoder;
use preserves::value::encoder::encode_bytes;
use preserves::value::IOValue;
use std::iter::Iterator;

mod samples;
use samples::*;

fn decode_all<'de>(bytes: &'de [u8]) -> Result<Vec<IOValue>, std::io::Error> {
    let d = decoder::from_bytes(bytes);
    d.collect()
}

#[test] fn run() -> std::io::Result<()> {
    let mut fh = std::fs::File::open("../../tests/samples.bin").unwrap();
    let mut d = decoder::from_read(&mut fh);
    let tests: TestCases = deserialize_from_value(&d.next().unwrap().unwrap()).unwrap();
    // println!("{:#?}", tests);

    for (Symbol(ref name), ref case) in tests.tests {
        println!("{:?} ==> {:?}", name, case);
        match case {
            TestCase::Test(ref bin, ref val) => {
                assert_eq!(&decode_all(&encode_bytes(val)?[..])?, &[val.clone()]);
                assert_eq!(&decode_all(&bin[..])?, &[val.clone()]);
                assert_eq!(&encode_bytes(val)?, bin);
            }
            TestCase::NondeterministicTest(ref bin, ref val) => {
                // The test cases in samples.txt are carefully
                // written so that while strictly
                // "nondeterministic", the order of keys in
                // dictionaries follows Preserves order.
                assert_eq!(&decode_all(&encode_bytes(val)?[..])?, &[val.clone()]);
                assert_eq!(&decode_all(&bin[..])?, &[val.clone()]);
                assert_eq!(&encode_bytes(val)?, bin);
            }
            TestCase::StreamingTest(ref bin, ref val) => {
                assert_eq!(&decode_all(&encode_bytes(val)?[..])?, &[val.clone()]);
                assert_eq!(&decode_all(&bin[..])?, &[val.clone()]);
            }
            TestCase::DecodeTest(ref bin, ref val) => {
                assert_eq!(&decode_all(&encode_bytes(val)?[..])?, &[val.clone()]);
                assert_eq!(&decode_all(&bin[..])?, &[val.clone()]);
            }
            TestCase::ParseError(_) => (),
            TestCase::ParseShort(_) => (),
            TestCase::ParseEOF(_) => (),
            TestCase::DecodeError(ref bin) => {
                match decode_all(&bin[..]) {
                    Ok(_) => panic!("Unexpected success"),
                    Err(e) => if is_syntax_io_error(&e) {
                        ()
                    } else {
                        panic!("Unexpected error {:?}", e)
                    }
                }
            }
            TestCase::DecodeShort(ref bin) => {
                assert!(if let Err(e) = decoder::from_bytes(bin).next().unwrap() {
                    is_eof_io_error(&e)
                } else {
                    false
                })
            }
            TestCase::DecodeEOF(ref bin) => {
                assert!(decoder::from_bytes(bin).next().is_none());
            }
        }
    }
    Ok(())
}
